/*
 * (c) Copyright 2020 CORSIKA Project, corsika-project@lists.kit.edu
 *
 * See file AUTHORS for a list of contributors.
 *
 * This software is distributed under the terms of the GNU General Public
 * Licence version 3 (GPL Version 3). See file LICENSE for a full version of
 * the license.
 */

#include <corsika/process/qgsjetII/Interaction.h>

#include <corsika/environment/Environment.h>
#include <corsika/environment/NuclearComposition.h>
#include <corsika/geometry/FourVector.h>
#include <corsika/geometry/QuantityVector.h>
#include <corsika/process/qgsjetII/QGSJetIIFragmentsStack.h>
#include <corsika/process/qgsjetII/QGSJetIIStack.h>
#include <corsika/process/qgsjetII/qgsjet-II-04.h>
#include <corsika/setup/SetupStack.h>
#include <corsika/setup/SetupTrajectory.h>
#include <corsika/utl/COMBoost.h>

#include <random>
#include <sstream>
#include <string>
#include <tuple>

using std::cout;
using std::endl;
using std::ostringstream;
using std::string;
using std::tuple;

using namespace corsika;
using namespace corsika::setup;
using SetupParticle = setup::Stack::StackIterator;
using SetupProjectile = setup::StackView::StackIterator;
using Track = Trajectory;

namespace corsika::process::qgsjetII {

  Interaction::Interaction(const string& dataPath)
      : data_path_(dataPath) {
    if (dataPath == "") {
      if (std::getenv("CORSIKA_DATA")) {
        data_path_ = string(std::getenv("CORSIKA_DATA")) + "/QGSJetII/";
        cout << "Searching for QGSJetII data tables in " << data_path_ << endl;
      }
    }
  }

  Interaction::~Interaction() { cout << "QgsjetII::Interaction n=" << count_ << endl; }

  void Interaction::Init() {
    // initialize QgsjetII
    if (!initialized_) {
      qgset_();
      datadir DIR(data_path_);
      qgaini_(DIR.data);
      initialized_ = true;
    }
  }

  units::si::CrossSectionType Interaction::GetCrossSection(
      const particles::Code beamId, const particles::Code targetId,
      const units::si::HEPEnergyType Elab, const unsigned int Abeam,
      const unsigned int targetA) const {
    using namespace units::si;
    double sigProd = std::numeric_limits<double>::infinity();

    if (process::qgsjetII::CanInteract(beamId)) {

      const int xsCode = process::qgsjetII::GetQgsjetIIXSCodeRaw(beamId);
      int iTarget = 1;
      if (particles::IsNucleus(targetId)) {
        iTarget = targetA;
        if (iTarget > maxMassNumber_ || iTarget <= 0) {
          std::ostringstream txt;
          txt << "QgsjetII target outside range. iTarget=" << iTarget;
          throw std::runtime_error(txt.str().c_str());
        }
      }
      int iProjectile = 1;
      if (particles::IsNucleus(beamId)) {
        iProjectile = Abeam;
        if (iProjectile > maxMassNumber_ || iProjectile <= 0)
          throw std::runtime_error("QgsjetII target outside range. ");
      }

      cout << "QgsjetII::GetCrossSection Elab=" << Elab << " xs-code=" << xsCode
           << " iProjectile=" << iProjectile << " iTarget=" << iTarget << endl;
      sigProd = qgsect_(Elab / 1_GeV, xsCode, iProjectile, iTarget);
      cout << "QgsjetII::GetCrossSection sigProd=" << sigProd << endl;
    }

    return sigProd * 1_mb;
  }

  template <>
  units::si::GrammageType Interaction::GetInteractionLength(
      SetupParticle const& vP) const {

    using namespace units;
    using namespace units::si;
    using namespace geometry;

    // coordinate system, get global frame of reference
    CoordinateSystem& rootCS =
        RootCoordinateSystem::GetInstance().GetRootCoordinateSystem();

    const particles::Code corsikaBeamId = vP.GetPID();

    // beam particles for qgsjetII : 1, 2, 3 for p, pi, k
    // read from cross section code table
    const bool kInteraction = process::qgsjetII::CanInteract(corsikaBeamId);

    // FOR NOW: assume target is at rest
    MomentumVector pTarget(rootCS, {0_GeV, 0_GeV, 0_GeV});

    // total momentum and energy
    HEPEnergyType Elab = vP.GetEnergy();

    cout << "Interaction: LambdaInt: \n"
         << " input energy: " << vP.GetEnergy() / 1_GeV << endl
         << " beam can interact:" << kInteraction << endl
         << " beam pid:" << vP.GetPID() << endl;

    if (kInteraction) {

      int Abeam = 0;
      if (particles::IsNucleus(vP.GetPID())) Abeam = vP.GetNuclearA();

      // get target from environment
      /*
        the target should be defined by the Environment,
        ideally as full particle object so that the four momenta
        and the boosts can be defined..
      */

      auto const* currentNode = vP.GetNode();
      const auto& mediumComposition =
          currentNode->GetModelProperties().GetNuclearComposition();

      si::CrossSectionType weightedProdCrossSection = mediumComposition.WeightedSum(
          [=](particles::Code targetID) -> si::CrossSectionType {
            int targetA = 0;
            if (corsika::particles::IsNucleus(targetID))
              targetA = particles::GetNucleusA(targetID);
            return GetCrossSection(corsikaBeamId, targetID, Elab, Abeam, targetA);
          });

      cout << "Interaction: "
           << "IntLength: weighted CrossSection (mb): " << weightedProdCrossSection / 1_mb
           << endl;

      // calculate interaction length in medium
      GrammageType const int_length = mediumComposition.GetAverageMassNumber() *
                                      units::constants::u / weightedProdCrossSection;
      cout << "Interaction: "
           << "interaction length (g/cm2): " << int_length / (0.001_kg) * 1_cm * 1_cm
           << endl;

      return int_length;
    }

    return std::numeric_limits<double>::infinity() * 1_g / (1_cm * 1_cm);
  }

  /**
     In this function QGSJETII is called to produce one event. The
     event is copied (and boosted) into the shower lab frame.
   */

  template <>
  process::EProcessReturn Interaction::DoInteraction(SetupProjectile& vP) {

    using namespace units;
    using namespace utl;
    using namespace units::si;
    using namespace geometry;

    const auto corsikaBeamId = vP.GetPID();
    cout << "ProcessQgsjetII: "
         << "DoInteraction: " << corsikaBeamId << " interaction? "
         << process::qgsjetII::CanInteract(corsikaBeamId) << endl;

    if (process::qgsjetII::CanInteract(corsikaBeamId)) {

      const CoordinateSystem& rootCS =
          RootCoordinateSystem::GetInstance().GetRootCoordinateSystem();

      // position and time of interaction, not used in QgsjetII
      Point pOrig = vP.GetPosition();
      TimeType tOrig = vP.GetTime();

      // define target
      // for QgsjetII is always a single nucleon
      // FOR NOW: target is always at rest
      const auto targetEnergyLab = 0_GeV + constants::nucleonMass;
      const auto targetMomentumLab = MomentumVector(rootCS, 0_GeV, 0_GeV, 0_GeV);
      const FourVector PtargLab(targetEnergyLab, targetMomentumLab);

      // define projectile
      HEPEnergyType const projectileEnergyLab = vP.GetEnergy();
      auto const projectileMomentumLab = vP.GetMomentum();

      int beamA = 0;
      if (particles::IsNucleus(corsikaBeamId)) beamA = vP.GetNuclearA();

      const HEPEnergyType projectileEnergyLabPerNucleon =
          particles::IsNucleus(corsikaBeamId) ? projectileEnergyLab / beamA
                                              : projectileEnergyLab;

      cout << "Interaction: ebeam lab: " << projectileEnergyLab / 1_GeV << endl
           << "Interaction: pbeam lab: " << projectileMomentumLab.GetComponents() / 1_GeV
           << endl;
      cout << "Interaction: etarget lab: " << targetEnergyLab / 1_GeV << endl
           << "Interaction: ptarget lab: " << targetMomentumLab.GetComponents() / 1_GeV
           << endl;

      cout << "Interaction: position of interaction: " << pOrig.GetCoordinates() << endl;
      cout << "Interaction: time: " << tOrig << endl;

      // sample target mass number
      auto const* currentNode = vP.GetNode();
      auto const& mediumComposition =
          currentNode->GetModelProperties().GetNuclearComposition();
      // get cross sections for target materials
      /*
        Here we read the cross section from the interaction model again,
        should be passed from GetInteractionLength if possible
       */
      auto const& compVec = mediumComposition.GetComponents();
      std::vector<si::CrossSectionType> cross_section_of_components(compVec.size());

      for (size_t i = 0; i < compVec.size(); ++i) {
        auto const targetId = compVec[i];
        int targetA = 0;
        if (corsika::particles::IsNucleus(targetId))
          targetA = particles::GetNucleusA(targetId);
        const auto sigProd =
            GetCrossSection(corsikaBeamId, targetId, projectileEnergyLab, beamA, targetA);
        cross_section_of_components[i] = sigProd;
      }

      const auto targetCode =
          mediumComposition.SampleTarget(cross_section_of_components, rng_);
      cout << "Interaction: target selected: " << targetCode << endl;

      int targetMassNumber = 1;               // proton
      if (particles::IsNucleus(targetCode)) { // nucleus
        targetMassNumber = particles::GetNucleusA(targetCode);
        if (targetMassNumber > maxMassNumber_)
          throw std::runtime_error("QgsjetII target mass outside range.");
      } else {
        if (targetCode != particles::Proton::GetCode())
          throw std::runtime_error("QgsjetII Taget not possible.");
      }
      cout << "Interaction: target qgsjetII code/A: " << targetMassNumber << endl;

      int projectileMassNumber = 1; // "1" means "hadron"
      QgsjetIIHadronType qgsjet_hadron_type =
          process::qgsjetII::GetQgsjetIIHadronType(corsikaBeamId);
      if (qgsjet_hadron_type == QgsjetIIHadronType::NucleusType) {
        projectileMassNumber = vP.GetNuclearA();
        if (projectileMassNumber > maxMassNumber_)
          throw std::runtime_error("QgsjetII projectile mass outside range.");
        std::array<QgsjetIIHadronType, 2> constexpr nucleons = {
            QgsjetIIHadronType::ProtonType, QgsjetIIHadronType::NeutronType};
        std::uniform_int_distribution select(0, 1);
        qgsjet_hadron_type = nucleons[select(rng_)];
      } else {
        // from conex: replace pi0 or rho0 with pi+/pi- in alternating sequence
        if (qgsjet_hadron_type == QgsjetIIHadronType::NeutralLightMesonType) {
          qgsjet_hadron_type = alternate_;
          alternate_ = (alternate_ == QgsjetIIHadronType::PiPlusType
                            ? QgsjetIIHadronType::PiMinusType
                            : QgsjetIIHadronType::PiPlusType);
        }
      }
      cout << "Interaction: projectile qgsjetII code/A: " << projectileMassNumber << " "
           << corsikaBeamId << endl;

      int qgsjet_hadron_type_int = static_cast<QgsjetIICodeIntType>(qgsjet_hadron_type);

      cout << "Interaction: "
           << " DoInteraction: E(GeV):" << projectileEnergyLab / 1_GeV << endl;
      count_++;
      qgini_(projectileEnergyLab / 1_GeV, qgsjet_hadron_type_int, projectileMassNumber,
             targetMassNumber);
      qgconf_();

      // bookkeeping
      MomentumVector Plab_final(rootCS, {0.0_GeV, 0.0_GeV, 0.0_GeV});
      HEPEnergyType Elab_final = 0_GeV;

      // to read the secondaries
      // define rotation to and from CoM frame
      // CoM frame definition in QgsjetII projectile: +z
      auto const& originalCS = projectileMomentumLab.GetCoordinateSystem();
      geometry::CoordinateSystem const zAxisFrame =
          originalCS.RotateToZ(projectileMomentumLab);

      // nuclear projectile fragments
      QGSJetIIFragmentsStack qfs;
      for (auto& fragm : qfs) {
        particles::Code idFragm = particles::Code::Nucleus;
        const int A = fragm.GetFragmentSize();
        int Z = 0;
        switch (A) {
          case 1: { // proton/neutron
            std::uniform_real_distribution<double> select;
            if (select(rng_) > 0.5) {
              idFragm = particles::Code::Proton;
              Z = 1;
            } else {
              idFragm = particles::Code::Neutron;
              Z = 0;
            }

            const HEPMassType projectileMass = particles::GetMass(idFragm);

            auto momentum = geometry::Vector(
                zAxisFrame, corsika::geometry::QuantityVector<hepmomentum_d>{
                                0.0_GeV, 0.0_GeV,
                                sqrt((projectileEnergyLab + projectileMass) *
                                     (projectileEnergyLab - projectileMass))});

            auto const energy = sqrt(momentum.squaredNorm() + square(projectileMass));
            momentum.rebase(originalCS); // transform back into standard lab frame
            std::cout << "secondary fragment> id=" << idFragm
                      << " p=" << momentum.GetComponents() << std::endl;
            auto pnew = vP.AddSecondary(
                tuple<particles::Code, units::si::HEPEnergyType, stack::MomentumVector,
                      geometry::Point, units::si::TimeType>{idFragm, energy, momentum,
                                                            pOrig, tOrig});
            Plab_final += pnew.GetMomentum();
            Elab_final += pnew.GetEnergy();
          } break;
          case 2: // deuterium
            Z = 1;
            break;
          case 3: // tritium
            Z = 1;
            break;
          case 4: // helium
            Z = 2;
            break;
          default: // nucleus
          {
            Z = int(A / 2.15 + 0.7);
          }
        }

        if (idFragm == particles::Code::Nucleus) {
          const HEPMassType nucleusMass =
              particles::Proton::GetMass() * Z + particles::Neutron::GetMass() * (A - Z);
          auto momentum = geometry::Vector(
              zAxisFrame, geometry::QuantityVector<hepmomentum_d>{
                              0.0_GeV, 0.0_GeV,
                              sqrt((projectileEnergyLabPerNucleon * A + nucleusMass) *
                                   (projectileEnergyLabPerNucleon * A - nucleusMass))});

          auto const energy = sqrt(momentum.squaredNorm() + square(nucleusMass));
          momentum.rebase(originalCS); // transform back into standard lab frame
          std::cout << "secondary fragment> id=" << idFragm
                    << " p=" << momentum.GetComponents() << " A=" << A << " Z=" << Z
                    << std::endl;
          auto pnew = vP.AddSecondary(
              tuple<particles::Code, units::si::HEPEnergyType, stack::MomentumVector,
                    geometry::Point, units::si::TimeType, unsigned short, unsigned short>{
                  idFragm, energy, momentum, pOrig, tOrig, A, Z});
          Plab_final += pnew.GetMomentum();
          Elab_final += pnew.GetEnergy();
        }
      }

      // secondaries
      QGSJetIIStack qs;
      for (auto& psec : qs) {

        auto momentum = psec.GetMomentum(zAxisFrame);
        auto const energy = psec.GetEnergy();

        momentum.rebase(originalCS); // transform back into standard lab frame
        std::cout << "secondary fragment> id="
                  << process::qgsjetII::ConvertFromQgsjetII(psec.GetPID())
                  << " p=" << momentum.GetComponents() << std::endl;
        auto pnew = vP.AddSecondary(
            tuple<particles::Code, units::si::HEPEnergyType, stack::MomentumVector,
                  geometry::Point, units::si::TimeType>{
                process::qgsjetII::ConvertFromQgsjetII(psec.GetPID()), energy, momentum,
                pOrig, tOrig});
        Plab_final += pnew.GetMomentum();
        Elab_final += pnew.GetEnergy();
      }
      cout << "conservation (all GeV): Ecm_final= n/a" /* << Ecm_final / 1_GeV*/ << endl
           << "Elab_final=" << Elab_final / 1_GeV
           << ", Plab_final=" << (Plab_final / 1_GeV).GetComponents()
           << ", N_wounded,targ="
           << QGSJetIIFragmentsStackData::GetWoundedNucleonsTarget()
           << ", N_wounded,proj="
           << QGSJetIIFragmentsStackData::GetWoundedNucleonsProjectile()
           << ", N_fragm,proj=" << qfs.GetSize() << endl;
    }
    return process::EProcessReturn::eOk;
  }

} // namespace corsika::process::qgsjetII
